"""Module containing common functionality for preparing template parameters."""

from typing import cast

import jinja2
from cargo2rpm.metadata import FeatureFlags, Package

from rust2rpm import log
from rust2rpm.conf import TomlConf
from rust2rpm.conf.toml import TestsComments, TestsSkip, TestsSkipExact, conf_comments_to_spec_comments
from rust2rpm.metadata import package_uses_rust_1_60_feature_syntax

MAX_SUMMARY_WIDTH = 80

RUST_PACKAGING_DEPS: dict[int, str] = {
    0: "rust-toolset",
    21: "rust-packaging >= 21",
    23: "rust-packaging >= 23",
    24: "cargo-rpm-macros >= 24",
    25: "cargo-rpm-macros >= 25",
    26: "cargo-rpm-macros >= 26",
}


RUST_PACKAGING_TARGET_MIN: dict[str, int] = {
    "fedora": 24,
    "mageia": 21,
    "opensuse": 21,
    "plain": 21,
}


def min_rust_packaging_dep(  # noqa: PLR0913
    package: Package,
    target: str,
    vendor_tarball: str | None,
    *,
    is_bin: bool,
    is_cdylib: bool,
    cargo_install_lib: bool,
    cargo_install_bin: bool,
) -> int:
    """Compute the minimum version of rust-packaging supported by this package."""
    if (not cargo_install_lib) or (not cargo_install_bin):
        min_dep = 26
    elif vendor_tarball:
        min_dep = 25
    elif package_uses_rust_1_60_feature_syntax(package.features):
        min_dep = 24
    elif is_bin or is_cdylib:
        min_dep = 23
    else:
        min_dep = 21

    return max((min_dep, RUST_PACKAGING_TARGET_MIN[target]))


def normalize_spdx_expr(expr: str) -> str:
    """Normalze SPDX expression (i.e. replace deprecated "/" operator with "OR")."""
    if "/" in expr:
        log.info("Replacing deprecated '/' operator in SPDX license expression with 'OR'")
        parts = [part.strip() for part in expr.split("/")]
        return " OR ".join(parts)
    return expr


def cargo_args_from_flags(
    feature_flags: FeatureFlags,
    required_features: set[str],
    enabled_by_default: set[str],
) -> str:
    """Format cargo arguments for the Rust macros from rust-packaging.

    Arguments:
        feature_flags: Feature flags from configuration.
        required_features: Features required by binary targets.
        enabled_by_default: Transitive closure of the "default" feature.

    Returns:
        Formatted string with cargo arguments.

    """
    if feature_flags.all_features:
        cargo_args = " -a"

    elif required_features:
        # remove required features that are already part of the enabled feature set
        for f in enabled_by_default:
            if f in required_features:
                required_features.remove(f)

        # merge, de-duplicate, and re-sort lists of enabled features
        if features_enable := feature_flags.features:
            enabled_features = sorted(set.union(set(required_features), set(features_enable)))
        else:
            enabled_features = sorted(required_features)

        cargo_args_list = ["-n"] if feature_flags.no_default_features else []

        if len(enabled_features) > 0:
            cargo_args_list.append(f"-f {','.join(enabled_features)}")

        cargo_args = " " + " ".join(cargo_args_list) if cargo_args_list else ""

    elif feature_flags.features:
        # list of features is already sorted when parsing the configuration file
        cargo_args_list = ["-n"] if feature_flags.no_default_features else []

        cargo_args_list.append(f"-f {','.join(feature_flags.features)}")
        cargo_args = " " + " ".join(cargo_args_list)

    elif feature_flags.no_default_features:
        cargo_args = " -n"

    else:
        cargo_args = ""

    return cargo_args


def make_rpm_summary(tomlconf: TomlConf, package: Package) -> str:
    """Collect RPM summary from settings or crate metadata and print warnings if it's too long."""
    if rpm_summary := tomlconf.package.summary:
        if len(rpm_summary) > MAX_SUMMARY_WIDTH:
            log.warn(
                f"The package Summary specified in rust2rpm.toml is too long (>{MAX_SUMMARY_WIDTH} characters).",
            )
    elif rpm_summary := package.get_summary():
        if len(rpm_summary) > MAX_SUMMARY_WIDTH:
            log.warn(
                f"Heuristics for generating a short Summary failed. The generated Summary string "
                f"is too long (>{MAX_SUMMARY_WIDTH} characters) and needs to be shortened "
                f"manually. This setting can be persisted with a 'package.summary' setting in "
                f"'rust2rpm.toml'.",
            )
    else:
        rpm_summary = "# FIXME"

    return rpm_summary


def spec_file_template(filename: str) -> jinja2.Template:
    """Load Jinja2 spec file template."""
    env = jinja2.Environment(
        loader=jinja2.ChoiceLoader([jinja2.FileSystemLoader(["/"]), jinja2.PackageLoader("rust2rpm", "templates")]),
        extensions=["jinja2.ext.do"],
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=False,
        undefined=jinja2.StrictUndefined,
        keep_trailing_newline=True,
    )

    return env.get_template(filename)


def conf_to_bcond_check(conf: TomlConf) -> int:
    """Compute value of the "check" bcond based on the rust2rpm configuration."""
    if tests_run := conf.tests.run:
        return 0 if "none" in tests_run else 1
    return 1


def conf_to_feature_flags(conf: TomlConf) -> FeatureFlags:
    """Compute feature flags based on the rust2rpm configuration."""
    return FeatureFlags(
        all_features=conf.features.enable_all is True,
        no_default_features=conf.features.disable_default is True,
        features=sorted(set(conf.features.enable) if conf.features.enable else set()),
    )


def conf_to_bcond_check_comments(conf: TomlConf) -> list[str]:
    """Compute comments associated with "%bcond check" based on the rust2rpm configuration."""
    tests_run = [conf.tests.run] if isinstance(conf.tests.run, str) else conf.tests.run

    if tests_run == ["none"]:
        return conf_comments_to_spec_comments(cast(list, conf.tests.comments))

    # the bcond only has associated comments if no tests are run
    return []


def conf_to_bin_rename_commands(conf: TomlConf) -> list[str]:
    """Compute commands for renaming executables based on the rust2rpm configuration."""
    return [
        f"mv %{{buildroot}}/%{{_bindir}}/{old} %{{buildroot}}/%{{_bindir}}/{new}"
        for old, new in conf.package.bin_renames.items()
    ]


def format_cargo_test_command(
    *,
    kind: str | None,
    cargo_args: str,
    skip_exact: bool,
    skip: list[str],
    comments: list[str],
) -> str:
    """Prepare a pre-formatted %cargo_test command.

    Arguments:
        kind: Test target kind ("lib", "bin", "doc", "tests", "bins").
        cargo_args: Pre-formatted arguments that are passed to all `%cargo_*`
            macros.
        skip_exact: Toggle inclusion of the `--exact` flag.
        skip: List of names of tests (or substrings of tests) that are
            `--skip`ped.
        comments: List of comments associated with this `%cargo_test` macro
            invocation.

    Returns:
        Pre-formatted `%cargo_test` command.

    """
    parts = []

    has_args = len(skip) > 0 or skip_exact

    if len(skip) <= 1:
        parts.append("%cargo_test")
        end = ""
    else:
        parts.append("%{cargo_test")
        end = "}"

    parts.append(cargo_args)

    if has_args or kind:
        parts.append(" --")

    if kind is not None:
        parts.append(f" --{kind}")

    if has_args:
        parts.append(" --")

    if skip_exact:
        parts.append(" --exact")

    if len(skip) == 1:
        parts.append(f" --skip {skip[0]}")
    elif len(skip) > 1:
        parts.append(" %{shrink:\n")
        parts.extend(f"    --skip {name}\n" for name in skip)
        parts.extend("}")

    parts.append(end)

    return "".join(line + "\n" for line in conf_comments_to_spec_comments(comments)) + "".join(parts)


def conf_to_cargo_test_commands(conf: TomlConf, cargo_args: str) -> list[str]:
    """Prepare a list of pre-formatted %cargo_test command based on the rust2rpm configuration.

    Arguments:
        conf: Global rust2rpm configuration.
        cargo_args: Pre-formatted arguments that are passed to all `%cargo_*`
            macros.

    Returns:
        List of pre-formatted `%cargo_test` commands.

    """
    tests_run = [conf.tests.run] if isinstance(conf.tests.run, str) else conf.tests.run

    if tests_run == ["none"]:
        return [
            format_cargo_test_command(
                kind=None,
                cargo_args=cargo_args,
                skip=[],
                skip_exact=False,
                comments=[],
            ),
        ]

    if len(tests_run) == 0 or tests_run == ["all"]:
        return [
            format_cargo_test_command(
                kind=None,
                cargo_args=cargo_args,
                skip=cast(list[str], conf.tests.skip),
                skip_exact=cast(bool, conf.tests.skip_exact),
                comments=cast(list[str], conf.tests.comments),
            ),
        ]

    commands = []

    if isinstance(conf.tests.comments, list):
        commands.extend(conf_comments_to_spec_comments(conf.tests.comments))

    for run in tests_run:
        kind = run
        skip = getattr(conf.tests.skip, kind) if isinstance(conf.tests.skip, TestsSkip) else conf.tests.skip
        skip_exact = (
            getattr(conf.tests.skip_exact, kind)
            if isinstance(conf.tests.skip_exact, TestsSkipExact)
            else conf.tests.skip_exact
        )
        comments = getattr(conf.tests.comments, kind) if isinstance(conf.tests.comments, TestsComments) else []

        commands.append(
            format_cargo_test_command(
                kind=kind,
                cargo_args=cargo_args,
                skip=skip,
                skip_exact=skip_exact,
                comments=comments,
            ),
        )

    return commands
