"""Module containing functionality for loading projects from "crate" archives."""

import shutil
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path

from cargo2rpm.metadata import Metadata, Package
from cargo2rpm.semver import Version, VersionReq

from rust2rpm import log
from rust2rpm.cli import Options
from rust2rpm.conf import TomlConf
from rust2rpm.cratesio import download_crate, query_available_versions, query_newest_version
from rust2rpm.inspect import files_from_crate
from rust2rpm.patching import make_patches
from rust2rpm.vendor import generate_vendor_tarball

from .meta import InvalidVersionError, PatchFile, Project
from .shared import file_name_auto_patch, file_name_manual_patch


class OfflineError(Exception):
    """Raised when attempting actions that would require internet access in offline mode."""


def parse_crate_file_name(path: str) -> tuple[str, str]:
    """Parse crate file name into crate name and crate version.

    Arguments:
        path: File name (including ".crate" suffix)

    Returns:
        Tuple consisting of (name, version) of the crate.

    """
    filename = path.removesuffix(".crate")

    try:
        name, version = filename.rsplit("-", 1)
        Version.parse(version)
    except ValueError:
        pass
    else:
        return name, version

    try:
        name, version, pre = filename.rsplit("-", 2)
        Version.parse(f"{version}-{pre}")
    except ValueError as exc:
        msg = f"Crate version cannot be determined from crate file name: {path}"
        raise InvalidVersionError(msg) from exc
    else:
        return name, version


def package_name_suffixed(name: str, suffix: str | None) -> str:
    """Construct RPM package name for a Rust crate with given name and suffix.

    This takes into account the rules for versioning compat packages from the
    Fedora Packaging Guidelines (i.e. the compat suffix must use a `_` character
    as a separator of the base package name ends in a digit).
    """
    joiner = "_" if suffix and name[-1].isdigit() else ""
    return "rust-" + name + joiner + (suffix or "")


def package_name_compat(name: str, version: str) -> str:
    """Construct RPM package name for a Rust crate with the correct version suffix for a "compat" package.

    To avoid dependency resolution issues that might be hard to debug, only one
    version of a crate may be packaged for every "major" release branch of a
    crate as defined by the cargo flavor of SemVer.
    """
    sv = Version.parse(version)
    if sv.major > 0:
        suffix = str(sv.major)
    elif sv.minor > 0:
        suffix = f"0.{sv.minor}"
    else:
        suffix = f"0.0.{sv.patch}"
    return package_name_suffixed(name, suffix)


def resolve_version(version: str, query: Callable[[], list[Version]]) -> Version | None:
    """Query the crates.io API to resolve a fully specified or partial version string.

    Arguments:
        version: Full or partial version string.
        query: Callable that takes zero arguments and returns the list of
            available versions for the given crate.

    Returns:
        Greatest available version matching the specified version string,
        or `None` if no available version matches the given version string.

    Raises:
        `InvalidVersionError` if the version string does not parse correctly
            as either a SemVer version or version requirement string.

    """
    try:
        return Version.parse(version)
    except ValueError:
        pass

    try:
        parsed_version = VersionReq.parse(version)
    except ValueError as exc:
        msg = f"Version argument does not parse as a version or version requirement: {version!r}"
        raise InvalidVersionError(msg) from exc

    log.info("Resolving version requirement ...")

    available_versions = query()
    filtered_versions = [*filter(lambda x: x in parsed_version, available_versions)]
    return max(filtered_versions) if filtered_versions else None


@dataclass(frozen=True)
class CrateProject(Project):
    """Description of a project based on a crate archive."""

    compat: bool
    """Flag that indicates whether this package is intended to be a compat package."""

    @property
    def package(self) -> Package:
        """Metadata for the single cargo "package" present in cargo metadata."""
        return self.metadata.packages[0]

    @property
    def rpm_name(self) -> str:
        """RPM source package name based on project metadata."""
        if self.compat:
            return package_name_compat(self.name, self.version)

        return package_name_suffixed(self.name, None)

    @staticmethod
    def from_crate_file(
        path: Path,
        name: str,
        version: str,
        out_dir: Path | None,
        conf: TomlConf,
        options: Options,
    ) -> "Project":
        """Load project metadata based on contents of a crate archive.

        Arguments:
            path: Path of the crate file.
            name: Name of the crate.
            version: Version of the crate.
            out_dir: Output directory for any generated or downloaded files.
            conf: Global rust2rpm configuration.
            options: Options and flags derived from command-line arguments.

        Returns:
            Project metadata based on the contents of the crate archive.

        Raises:
            `ValueError` when the crate archive contains multiple crates.

        """
        # process files from  a .crate archive
        with files_from_crate(path, name, version, conf) as (toml_path, license_files, doc_files):
            metadata = Metadata.from_cargo(str(toml_path))

            if len(metadata.packages) > 1:
                log.error("Attempting to process a .crate file which contains a cargo workspace.")
                log.error("This mode of operation is unusual and not supported by rust2rpm.")
                msg = "Failed to process invalid .crate file (cargo workspace)"
                raise ValueError(msg)

            package = metadata.packages[0]
            version = package.version

            reapply_path = (out_dir or Path()) / file_name_manual_patch(name) if options.reuse_patch else None

            auto_diff, manual_diff = make_patches(
                name,
                version,
                toml_path,
                reapply_path,
                options,
            )

            # ensure metadata is up-to-date with changes from patches
            metadata = Metadata.from_cargo(str(toml_path))

            vendor_tarball = generate_vendor_tarball(toml_path, name, version, out_dir or Path(), options.vendor)

        auto_patch = PatchFile(file_name_auto_patch(name), diff) if (diff := auto_diff) else None
        manual_patch = PatchFile(file_name_manual_patch(name), diff) if (diff := manual_diff) else None

        return CrateProject(
            name,
            version,
            metadata,
            license_files,
            doc_files,
            auto_patch,
            manual_patch,
            vendor_tarball,
            options.compat,
        )

    @staticmethod
    def from_local_crate(
        path: Path,
        out_dir: Path | None,
        conf: TomlConf,
        options: Options,
    ) -> "Project":
        """Load project metadata based on contents of a crate archive.

        This is a wrapper method around `CrateProject.from_crate_file` but
        it parses the file name into crate (name, version) so they can be
        passed as arguments and don't need to be explicitly specified in
        cases where they can be inferred from the file name.

        Arguments:
            path: Path of the crate file.
            out_dir: Output directory for any generated or downloaded files.
            conf: Global rust2rpm configuration.
            options: Options and flags derived from command-line arguments.

        Returns:
            Project metadata based on the contents of the crate archive.

        Raises:
            `ValueError` when the crate archive contains multiple crates.

        """
        # determine name and version from the filename
        name, version = parse_crate_file_name(path.name)

        return CrateProject.from_crate_file(path, name, version, out_dir, conf, options)

    @staticmethod
    def from_crates_io(
        name: str,
        version: str | None,
        out_dir: Path | None,
        conf: TomlConf,
        options: Options,
    ) -> "Project":  # pragma nocover: requires internet access
        """Load project metadata based on contents of a crate archive.

        This method downloads the given crate from the crates.io registry
        and then calls `CrateProject.from_crate_file` internally.

        The behaviour which version from crates.io is used depends on the value
        of the `version` argument:

        - If the argument is `None`, query crates.io for the greatest
            non-prerelease version and use it. If there are no non-yanked
            non-prerelease versions, `InvalidVersionError` is raised.
        - If the argument is a full SemVer version string (i.e.
            `major.minor.patch`), then it is used to query crates.io for this
            exact version. If the version does not exit, `InvalidVersionError`
            is raised.
        - If the argument is a SemVer version requirement (i.e. operator plus
            partial version string), then crates.io is queried for available
            versions, and the greatest version that matches the given
            requirement is used. If no version matches the requirement,
            `InvalidVersionError` is raised.

        In "offline" mode, the version needs to be fully specified since
        crates.io cannot be queried for a list of available versions to choose
        from. The specified version needs to have been previously downloaded
        into the rust2rpm download cache for "offline" mode to work.

        Arguments:
            name: Name of the crate in the crates.io registry.
            version: Optional version or version requirement string, used
                for filtering and choosing a version from all available versions.
            out_dir: Output directory for any generated or downloaded files.
            conf: Global rust2rpm configuration.
            options: Options and flags derived from command-line arguments.

        Returns:
            Project metadata based on the contents of the crate archive.

        Raises:
            `InvalidVersionError` is raised if the `version` argument is invalid
                or if there is no available version that matches the argument.
            `OfflineError` is raised if the version argument was not specific
                enough or if the specified crate version had not been downloaded
                yet.

        """
        if version:
            if options.offline:
                try:
                    valid_version = Version.parse(version)
                except ValueError as exc:
                    msg = "Crate version needs to be explicitly and fully specified in offline mode."
                    raise OfflineError(msg) from exc

                crate_file_path = download_crate(name, valid_version, offline=True)
                version = str(valid_version)

            else:
                # version or partial version was specified
                resolved_version = resolve_version(version, lambda: query_available_versions(name))

                if resolved_version:
                    log.info(f"Partial version matched with available version: {resolved_version}")
                else:
                    msg = f"Partial version did not match any available version: {version}"
                    raise InvalidVersionError(msg)

                crate_file_path = download_crate(name, resolved_version, offline=False)
                version = str(resolved_version)

        else:
            if options.offline:
                msg = "Crate version needs to be explicitly and fully specified in offline mode."
                raise OfflineError(msg)

            # no version was specified: download latest
            newest_version = query_newest_version(name)

            crate_file_path = download_crate(name, newest_version, offline=False)
            version = str(newest_version)

        copy_target = (out_dir or Path.cwd()) / crate_file_path.name
        copy_target.parent.mkdir(parents=True, exist_ok=True)

        if options.offline and not crate_file_path.exists() and not copy_target.exists():
            msg = "The specified crate version has not been downloaded yet."
            raise OfflineError(msg)

        if options.store_crate:
            # crate file was manually downloaded but is not in the cache yet
            if options.offline and copy_target.exists() and not crate_file_path.exists():
                shutil.copy2(copy_target, crate_file_path)

            # crate file was downloaded to the cache but not yet copied to CWD
            if not (copy_target.exists() and crate_file_path.samefile(copy_target)):
                shutil.copy2(crate_file_path, copy_target)

        return CrateProject.from_crate_file(crate_file_path, name, version, out_dir, conf, options)
